# Import a export CSV

# Jak vypadá CSV vypadlé z novějších verzí Excelu:
#   - je v UTF-8 a má na začátku BOM
#   - řádky ukončuje CR+LF
#   - separátor je čárka
#   - quoting pomocí uvozovek

import csv
import dataclasses
from typing import Type, List, IO, Sequence


class TSV(csv.Dialect):
    delimiter = '\t'
    doublequote = True
    lineterminator = '\n'
    quotechar = '"'
    quoting = csv.QUOTE_MINIMAL


csv.register_dialect('tsv', TSV)


class Row:
    """Řádek tabulky je dataclass, která je potomkem této třídy. Atributy
    odpovídají sloupečkům, musí být typu str a mít defaultní hodnotu."""


def write(file: IO, dialect: str, row_class: Type[Row], rows: Sequence[Row]):
    writer = csv.writer(file, dialect=dialect)

    columns = [field.name for field in dataclasses.fields(row_class)]
    writer.writerow(columns)

    for row in rows:
        r = [getattr(row, c) for c in columns]
        writer.writerow(r)


def read(file: IO, dialect: str, row_class: Type[Row]):
    reader = csv.reader(file, dialect=dialect, strict=True)

    header: List[str] = []
    rows: List[Row] = []
    columns = set(field.name for field in dataclasses.fields(row_class))

    for r in reader:
        if reader.line_num == 1:
            if r[0].startswith('\ufeff'):
                # Excel's BOM
                r[0] = r[0][1:]
            header = r
        else:
            row = row_class()
            for i in range(min(len(r), len(header))):
                f = header[i]
                x = r[i].strip()
                if f in columns and x != "" and x != '-':
                    setattr(row, f, x)
            rows.append(row)

    return rows
