from dataclasses import dataclass
import io
import re
from typing import List, Optional, Any, Dict, Callable, Type, TypeVar

import mo.csv
import mo.db as db
import mo.rights
import mo.users
import mo.util
from mo.util import logger

RowType = TypeVar('RowType', bound=mo.csv.Row)


@dataclass
class ContestImportRow(mo.csv.Row):
    email: str = ""
    krestni: str = ""
    prijmeni: str = ""
    kod_skoly: str = ""
    rocnik: str = ""
    rok_naroz: str = ""
    kod_mista: str = ""
    kod_oblasti: str = ""


@dataclass
class ProctorImportRow(mo.csv.Row):
    email: str = ""
    krestni: str = ""
    prijmeni: str = ""
    kod_mista: str = ""


class Import:
    # Výsledek importu
    errors: List[str]
    cnt_rows: int = 0
    cnt_new_users: int = 0
    cnt_new_participants: int = 0
    cnt_new_participations: int = 0
    cnt_new_roles: int = 0

    # Interní: Co zrovna importujeme
    user: db.User
    round: Optional[db.Round]
    contest: Optional[db.Contest]

    # Interní: Stav importu
    place_cache: Dict[str, db.Place]
    school_place_cache: Dict[str, db.Place]
    rr: Optional[mo.rights.Rights]
    place_rights_cache: Dict[int, bool]
    new_user_ids: List[int]
    line_number: int = 0

    def __init__(self, user: db.User):
        self.errors = []
        self.user = user
        self.rr = None
        self.place_cache = {}
        self.school_place_cache = {}
        self.place_rights_cache = {}
        self.new_user_ids = []

    def error(self, msg: str) -> Any:
        if self.line_number > 0:
            msg = f"Řádek {self.line_number}: {msg}"
        self.errors.append(msg)
        return None     # Kdyby bylo otypováno správně jako -> None, při volání by si mypy stěžoval

    def parse_email(self, email: str) -> Optional[str]:
        if email == "":
            return self.error('Chybí e-mailová adresa')

        if not re.fullmatch(r'.+@.+', email):
            return self.error('Chybný formát e-mailové adresy')

        if not re.fullmatch(r'[!-~]+', email):
            return self.error('E-mailová adresa obsahuje znaky mimo ASCII')

        return mo.users.normalize_email(email)

    def parse_name(self, name: str) -> Optional[str]:
        if name == "":
            return self.error('Jméno nesmí být prázdné')

        # XXX: Tato kontrola úmyslně není striktní, aby prošla i jména jako 'de Beer'
        if name == name.lower():
            return self.error('Ve jméně nejsou velká písmena')

        if name == name.upper():
            return self.error('Ve jméně nejsou malá písmena')

        return name

    def check_rights(self, place: db.Place) -> bool:
        if place.place_id in self.place_rights_cache:
            return self.place_rights_cache[place.place_id]

        if self.rr is None:
            self.rr = mo.rights.Rights(self.user)

        round = self.round
        assert round is not None
        self.rr.get_for(place, round.year, round.category, round.seq)
        result = self.rr.have_right(mo.rights.Right.manage_contest)
        self.place_rights_cache[place.place_id] = result

        return result

    def parse_opt_place(self, kod: str) -> Optional[db.Place]:
        if kod == "":
            return None

        if kod in self.place_cache:
            return self.place_cache[kod]

        place = db.get_place_by_code(kod)
        if not place:
            return self.error(f'Místo "{kod}" nenalezeno')

        if not self.check_rights(place):
            return self.error(f'K místu "{kod}" nemáte práva na správu soutěže')

        self.place_cache[kod] = place
        return place

    def parse_school(self, kod: str) -> Optional[db.Place]:
        if kod == "":
            return self.error('Škola je povinná')

        if kod in self.school_place_cache:
            return self.school_place_cache[kod]

        place = db.get_place_by_code(kod, fetch_school=True)
        if not place:
            return self.error(f'Škola "{kod}" nenalezena')

        if place.type != db.PlaceType.school:
            return self.error(f'Kód školy "{kod}" neodpovídá škole')

        self.school_place_cache[kod] = place
        return place

    def parse_grade(self, rocnik: str, school: Optional[db.School]) -> Optional[str]:
        if not school:
            return None

        if (school.is_ss and re.fullmatch(r'\d/\d', rocnik)
                or school.is_zs and re.fullmatch(r'\d', rocnik)):
            return rocnik

        return self.error('Ročník neodpovídá typu školy: pro základní je to číslice, pro střední číslice/číslice')

    def parse_born(self, rok: str) -> Optional[int]:
        if not re.fullmatch(r'\d{4}', rok):
            return self.error('Rok narození musí být čtyřciferné číslo')

        r = int(rok)
        if r < 2000 or r > 2099:
            return self.error('Rok narození musí být v intervalu [2000,2099]')

        return r

    def find_or_create_user(self, email: str, krestni: str, prijmeni: str) -> Optional[db.User]:
        sess = db.get_session()
        user = sess.query(db.User).filter_by(email=email).one_or_none()
        if user:
            if user.first_name != krestni or user.last_name != prijmeni:
                return self.error(f'Osoba již registrována s odlišným jménem {user.full_name()}')
        else:
            user = db.User(email=email, first_name=krestni, last_name=prijmeni)
            sess.add(user)
            sess.flush()    # Aby uživatel dostal user_id
            logger.info(f'Import: Založen uživatel user=#{user.user_id} email=<{user.email}>')
            mo.util.log(
                type=db.LogType.user,
                what=user.user_id,
                details={'action': 'import', 'new': db.row2dict(user)},
            )
            self.cnt_new_users += 1
            self.new_user_ids.append(user.user_id)
        return user

    def find_or_create_participant(self, user: db.User, year: int, school_id: int, birth_year: int, grade: str) -> Optional[db.Participant]:
        sess = db.get_session()
        part = sess.query(db.Participant).get((user.user_id, year))
        if part:
            if (part.school != school_id
                    or part.grade != grade
                    or part.birth_year != birth_year):
                return self.error('Účastník již zaregistrován s odlišnou školou/ročníkem/rokem narození')
        else:
            part = db.Participant(user=user, year=year, school=school_id, birth_year=birth_year, grade=grade)
            sess.add(part)
            logger.info(f'Import: Založen účastník #{user.user_id}')
            mo.util.log(
                type=db.LogType.participant,
                what=user.user_id,
                details={'action': 'import', 'new': db.row2dict(part)},
            )
            self.cnt_new_participants += 1

        return part

    def find_or_create_participation(self, user: db.User, contest: db.Contest, place: Optional[db.Place]) -> Optional[db.Participation]:
        if place is None:
            place = contest.place

        sess = db.get_session()
        pions = (sess.query(db.Participation)
                 .filter_by(user=user)
                 .filter(db.Participation.contest.has(db.Contest.round == contest.round))
                 .all())

        if not pions:
            pion = db.Participation(user=user, contest=contest, place_id=place.place_id, state=db.PartState.invited)
            sess.add(pion)
            logger.info(f'Import: Založena účast user=#{user.user_id} contest=#{contest.contest_id} place=#{place.place_id}')
            mo.util.log(
                type=db.LogType.participant,
                what=user.user_id,
                details={'action': 'add-to-contest', 'new': db.row2dict(pion)},
            )
            self.cnt_new_participations += 1
        elif len(pions) == 1:
            pion = pions[0]
            if pion.place != place:
                return self.error('Již se tohoto kola účastní v jiné oblasti')
        else:
            return self.error('Již se tohoto kola účastní ve vice oblastech, což by nemělo být možné')

        return pion

    def import_contest_row(self, r: ContestImportRow):
        num_prev_errs = len(self.errors)
        email = self.parse_email(r.email)
        krestni = self.parse_name(r.krestni)
        prijmeni = self.parse_name(r.prijmeni)
        school_place = self.parse_school(r.kod_skoly)
        rocnik = self.parse_grade(r.rocnik, (school_place.school if school_place else None))
        rok_naroz = self.parse_born(r.rok_naroz)
        misto = self.parse_opt_place(r.kod_mista)
        oblast = self.parse_opt_place(r.kod_oblasti)

        if (len(self.errors) > num_prev_errs
                or email is None
                or krestni is None
                or prijmeni is None
                or school_place is None
                or rocnik is None
                or rok_naroz is None):
            return

        user = self.find_or_create_user(email, krestni, prijmeni)
        if user is None:
            return

        part = self.find_or_create_participant(user, mo.current_year, school_place.place_id, rok_naroz, rocnik)
        if part is None:
            return

        if self.contest:
            contest = self.contest
            if oblast is not None and oblast.place_id != contest.place.place_id:
                return self.error('Účastník soutěží mimo oblast, do které se importuje')
        else:
            if oblast is None:
                return self.error('Je nutné uvést oblast')
            contest = db.get_session().query(db.Contest).filter_by(round=self.round, place=oblast).one_or_none()
            if contest is None:
                return self.error('V uvedené oblasti toto kolo neprobíhá')

        self.find_or_create_participation(user, contest, misto)

    def import_proctor_row(self, r: ProctorImportRow):
        num_prev_errs = len(self.errors)
        email = self.parse_email(r.email)
        krestni = self.parse_name(r.krestni)
        prijmeni = self.parse_name(r.prijmeni)
        misto = self.parse_opt_place(r.kod_mista)

        if misto is None:
            return self.error('Kód místa je povinné uvést')

        if (len(self.errors) > num_prev_errs
                or email is None
                or krestni is None
                or prijmeni is None):
            return

        user = self.find_or_create_user(email, krestni, prijmeni)
        if user is None:
            return

        sess = db.get_session()
        round = self.round
        assert round is not None

        if (sess.query(db.UserRole)
                .filter_by(user=user, place=misto, role=db.RoleType.dozor,
                           category=round.category, year=round.year, seq=round.seq)
                .with_for_update()
                .first()):
            pass
        else:
            ur = db.UserRole(user=user, place=misto, role=db.RoleType.dozor,
                             category=round.category, year=round.year, seq=round.seq)
            sess.add(ur)
            sess.flush()
            logger.info(f'Import: Dozor user=#{user.user_id} place=#{misto.place_id} user_role=#{ur.user_role_id}')
            mo.util.log(
                type=db.LogType.user_role,
                what=ur.user_role_id,
                details={'action': 'import', 'new': db.row2dict(ur)},
            )
            self.cnt_new_roles += 1

    def generic_import(self, path: str, row_class: Type[RowType], process_row: Callable[['Import', RowType], None]) -> bool:
        try:
            with open(path) as file:
                rows: List[RowType] = mo.csv.read(file=file, dialect='excel', row_class=row_class)
        except UnicodeDecodeError:
            return self.error('Soubor není v kódování UTF-8')
        except Exception as e:
            return self.error(f'Chybná struktura tabulky: {e}')

        self.line_number = 2
        for row in rows:
            self.cnt_rows += 1
            process_row(self, row)
            if len(self.errors) >= 100:
                self.errors.append('Import přerušen pro příliš mnoho chyb')
                break
            self.line_number += 1

        if self.errors:
            logger.info('Import: Rollback')
            return False
        else:
            logger.info(f'Import: Hotovo (rows={self.cnt_rows} users={self.cnt_new_users} p-ants={self.cnt_new_participants} p-ions={self.cnt_new_participations} roles={self.cnt_new_roles})')
            return True

    def import_contest(self, round: db.Round, contest: Optional[db.Contest], path: str) -> bool:
        self.round = round
        self.contest = contest
        logger.info(f'Import: Účastníci ze souboru {path}: uid={self.user.user_id} round=#{round.round_id}')
        if not self.generic_import(path, ContestImportRow, Import.import_contest_row):
            db.get_session().rollback()
            return False

            mo.util.log(
                type=db.LogType.contest,
                what=contest.contest_id,
                details={'action': 'import'}
            )
            db.get_session().commit()

        mo.util.log(
            type=db.LogType.round,
            what=round.round_id,
            details={'action': 'import'}
        )
        db.get_session().commit()

        self.notify_users()
        return True

    def import_proctors(self, round: db.Round, path: str) -> bool:
        self.round = round
        self.contest = None
        logger.info(f'Import: Dozor ze souboru {path}: uid={self.user.user_id}')
        if not self.generic_import(path, ProctorImportRow, Import.import_proctor_row):
            return False

        mo.util.log(
            type=db.LogType.round,
            what=round.round_id,
            details={'action': 'import-proctors'}
        )
        db.get_session().commit()

        self.notify_users()
        return True

    def notify_users(self):
        # Projde všechny uživatele a těm, kteří ještě nemají nastavené heslo,
        # ani nepožádali o jeho reset, pošle mail s odkazem na reset. Každého
        # uživatele zpracováváme ve zvlášť transakci, aby se po chybě neztratila
        # informace o tom, že už jsme nějaké maily rozeslali.

        sess = db.get_session()
        for uid in self.new_user_ids:
            u = sess.query(db.User).get(uid)
            if u and not u.password_hash and not u.reset_at:
                token = mo.users.ask_reset_password(u)
                sess.commit()
                mo.util.send_new_account_email(u, token)
            else:
                sess.rollback()


def generic_template(row_class: Type[RowType]) -> str:
    out = io.StringIO()
    mo.csv.write(file=out, dialect='excel', row_class=row_class, rows=[])
    return out.getvalue()


def contest_template() -> str:
    return generic_template(ContestImportRow)


def proctor_template() -> str:
    return generic_template(ProctorImportRow)
