import datetime
import multiprocessing
import os
import pikepdf
import secrets
import werkzeug.utils

import mo.db as db
from mo.util import logger


class SubmitException(RuntimeError):
    pass


class Submitter:
    submit_dir: str

    def __init__(self, instance_path: str = 'data'):
        self.submit_dir = os.path.join(instance_path, 'submits')

    def submit_paper(self, paper: db.Paper, tmpfile: str):
        logger.info(f'Submit: Zpracovávám file={tmpfile} for=#{paper.for_user_obj.user_id} by=#{paper.uploaded_by_obj.user_id} type={paper.type.name}')
        t_start = datetime.datetime.now()

        try:
            self._do_submit(paper, tmpfile)
            duration = (datetime.datetime.now() - t_start).total_seconds()
            logger.info(f'Submit: Hotovo: file={paper.file_name} pages={paper.pages} bytes={paper.bytes} time={duration:.3f}')
        except SubmitException as e:
            duration = (datetime.datetime.now() - t_start).total_seconds()
            logger.info(f'Submit: Chyba: {e} (time={duration:.3f})')
            raise

    def _create_file_name(self, paper: db.Paper) -> str:
        user_dir = os.path.join(str(paper.task.round.year), str(paper.for_user_obj.user_id))
        sub_user_dir = os.path.join(self.submit_dir, user_dir)
        os.makedirs(sub_user_dir, exist_ok=True)

        secure_task_code = werkzeug.utils.secure_filename(paper.task.code)
        while True:
            nonce = secrets.token_hex(8)
            file_name = f'{secure_task_code}-{paper.type.name[:3]}-{nonce}.pdf'
            if not os.path.lexists(os.path.join(sub_user_dir, file_name)):
                break
            logger.warning(f'Retrying file creation for {sub_user_dir}/{file_name}')

        return os.path.join(user_dir, file_name)

    def _do_submit(self, paper: db.Paper, tmpfile: str):
        # Zpracování PDF spustíme v samostatném procesu, aby bylo dostatečně oddělené
        # FIXME: Omezit paměť apod.
        pipe_rx, pipe_tx = multiprocessing.Pipe(duplex=False)
        proc = multiprocessing.Process(name='submit', target=Submitter._process_pdf, args=(tmpfile, pipe_tx))
        proc.start()
        pipe_tx.close()

        if not pipe_rx.poll(10):
            proc.terminate()
            proc.join()
            raise SubmitException('Timeout při zpracování PDF.')

        try:
            result = pipe_rx.recv()
        except EOFError:
            result = None
            proc.terminate()
        proc.join()

        assert proc.exitcode is not None
        if proc.exitcode != 0:
            raise SubmitException(f'Interní chyba při zpracování PDF: exit code {proc.exitcode}.')
        if not result:
            raise SubmitException('Interní chyba při zpracování PDF: EOF.')

        if 'error' in result:
            logger.info('Submit: PDF error: ' + result['error'])
            raise SubmitException('Soubor není korektní PDF.')

        paper.bytes = os.path.getsize(tmpfile)
        paper.pages = result['pages']
        paper.file_name = self._create_file_name(paper)

        # FIXME: fsync?
        dest = os.path.join(self.submit_dir, paper.file_name)
        os.rename(tmpfile, dest)

    # Zpracování PDF běží v samostatném procesu, výsledek pošle jako slovník rourou.
    def _process_pdf(tmpfile, pipe):
        try:
            with pikepdf.open(tmpfile) as pdf:
                pages = len(pdf.pages)
        except pikepdf.PdfError as e:
            pipe.send({
                "error": str(e),
            })
            return

        pipe.send({
            "pages": pages,
        })
