# Správa uživatelů

import bcrypt
import datetime
from typing import Optional

import mo.db as db
import mo.util
import mo.tokens


def normalize_email(email: str) -> str:
    # XXX: Striktně vzato, tohle není korektní, protože některé domény mohou
    # mít case-sensitive levou stranu adresy. Ale i na nich se prakticky nevyskytují
    # levé strany s velkými písmeny, zatímco uživatelé při psaní adres běžně velká
    # a malá písmena zaměňují. Menší zlo tedy je normalizovat na malá písmena.
    return email.lower()


def user_by_email(email: str) -> db.User:
    email = normalize_email(email)
    return db.get_session().query(db.User).filter_by(email=email).first()


def user_by_uid(uid: int) -> db.User:
    return db.get_session().query(db.User).get(uid)


def set_password(user: db.User, passwd: str):
    salt = bcrypt.gensalt()
    hashed = bcrypt.hashpw(passwd.encode('utf-8'), salt)
    user.password_hash = hashed.decode('us-ascii')
    user.reset_at = None
    user.last_login_at = datetime.datetime.now()

    mo.util.log(
        type=db.LogType.user,
        what=user.user_id,
        details={"action": "set-passwd"},
    )


def check_password(user: db.User, passwd: str):
    return user.password_hash is not None and \
        bcrypt.checkpw(passwd.encode('utf-8'), user.password_hash.encode('us-ascii'))


def login(user: db.User):
    user.last_login_at = datetime.datetime.now()
    user.reset_at = None


def ask_reset_password(user: db.User) -> str:
    user.reset_at = datetime.datetime.now()
    when = int(user.reset_at.timestamp())
    token = mo.tokens.sign_token([str(user.user_id), str(when)], 'reset')

    mo.util.log(
        type=db.LogType.user,
        what=user.user_id,
        details={'action': 'ask-reset'},
    )

    return token


def check_reset_password(token: str) -> Optional[db.User]:
    fields = mo.tokens.verify_token(token, 'reset')
    if not fields or len(fields) != 2:
        return None
    user = db.get_session().query(db.User).filter_by(user_id=int(fields[0])).first()

    if user.password_hash is None:
        reset_token_validity_time = datetime.timedelta(days=28)
    else:
        reset_token_validity_time = datetime.timedelta(hours=24)

    now = datetime.datetime.now().astimezone()

    if (user
            and user.reset_at is not None
            and fields[1] == str(int(user.reset_at.timestamp()))
            and now - user.reset_at < reset_token_validity_time):
        return user
    else:
        return None


def cancel_reset_password(user: db.User):
    user.reset_at = None
    mo.util.log(
        type=db.LogType.user,
        what=user.user_id,
        details={'action': 'cancel-reset'},
    )


def do_reset_password(user: db.User):
    user.reset_at = None
    mo.util.log(
        type=db.LogType.user,
        what=user.user_id,
        details={'action': 'do-reset'},
    )
