from flask import render_template, request, g, redirect, url_for, flash
from flask_wtf import FlaskForm
import flask_wtf.file
import os
import secrets
from sqlalchemy import or_, and_
from sqlalchemy.orm import joinedload
from typing import Optional
import werkzeug.exceptions
import wtforms
import wtforms.validators as validators

import mo.config as config
import mo.db as db
import mo.submit
import mo.util
from mo.util import logger
from mo.web import app
import mo.web.util


@app.route('/user')
def user_index():
    sess = db.get_session()

    pions = (sess.query(db.Participation, db.Contest, db.Round)
             .select_from(db.Participation)
             .join(db.Contest)
             .join(db.Round)
             .filter(db.Participation.user == g.user)
             .options(joinedload(db.Contest.place))
             .order_by(db.Round.year.desc(), db.Round.category, db.Round.seq)
             .all())

    return render_template(
        'user_index.html',
        pions=pions,
        db=db,  # kvůli hodnotám enumů
    )


def get_contest(id: int) -> db.Contest:
    contest = (db.get_session().query(db.Contest)
               .options(joinedload(db.Contest.place),
                        joinedload(db.Contest.round))
               .get(id))

    if not contest:
        raise werkzeug.exceptions.NotFound()

    # FIXME: Kontrolovat nějak pion.state?
    pion = (db.get_session().query(db.Participation)
            .filter_by(user=g.user, contest=contest)
            .one_or_none())
    if not pion:
        raise werkzeug.exceptions.Forbidden()

    return contest


def get_task(contest: db.Contest, id: int) -> db.Task:
    task = db.get_session().query(db.Task).get(id)

    # Nezapomeňme zkontrolovat, že úloha patří do soutěže :)
    if not task or task.round_id != contest.round_id:
        raise werkzeug.exceptions.NotFound()

    return task


@app.route('/user/contest/<int:id>/')
def user_contest(id: int):
    contest = get_contest(id)

    task_sols = (db.get_session().query(db.Task, db.Solution)
                 .select_from(db.Task)
                 .outerjoin(db.Solution, and_(db.Solution.task_id == db.Task.task_id, db.Solution.user == g.user))
                 .filter(db.Task.round == contest.round)
                 .options(joinedload(db.Solution.final_submit_obj),
                          joinedload(db.Solution.final_feedback_obj))
                 .order_by(db.Task.code)
                 .all())

    return render_template(
        'user_contest.html',
        contest=contest,
        task_sols=task_sols,
        max_submit_size=config.MAX_CONTENT_LENGTH,
        db=db,  # kvůli hodnotám enumů
    )


@app.route('/user/contest/<int:id>/task-statement/')
def user_task_statement(id: int):
    contest = get_contest(id)

    if not contest.round.task_statement_available():
        logger.warn(f'Účastník #{g.user.user_id} chce zadání, na které nemá právo')
        raise werkzeug.exceptions.Forbidden()

    return mo.web.util.send_task_statement(contest.round)


class SubmitForm(FlaskForm):
    file = flask_wtf.file.FileField("Soubor", validators=[flask_wtf.file.FileRequired()])
    note = wtforms.TextAreaField("Poznámka", description="Zde můžete něco vzkázat organizátorům soutěže.")
    submit = wtforms.SubmitField('Odevzdat')


@app.route('/user/contest/<int:contest_id>/task/<int:task_id>/', methods=('GET', 'POST'))
def user_contest_task(contest_id: int, task_id: int):
    contest = get_contest(contest_id)
    task = get_task(contest, task_id)
    sess = db.get_session()

    round = contest.round
    if round.state == db.RoundState.preparing or (round.state == db.RoundState.running and not round.ct_can_submit()):
        # Dokud se kolo připravuje nebo čeká na zveřejnění zadání, tak ani nezobrazujeme
        # stránku, abychom něco neprozradili jménem úlohy
        raise werkzeug.exceptions.Forbidden()

    form = SubmitForm()
    if round.ct_can_submit() and form.validate_on_submit():
        # FIXME: Tohle je pomalé, dělá se tu zbytečná další kopie dat.
        # Nicméně werkzeugu by měla jít podstrčit stream factory,
        # která bude vyrábět streamy rovnou uložené v našem tmp.
        tmp_name = secrets.token_hex(16)
        tmp_path = os.path.join(app.instance_path, 'tmp', tmp_name)
        form.file.data.save(tmp_path)

        paper = db.Paper(task=task, for_user_obj=g.user, uploaded_by_obj=g.user, type=db.PaperType.solution, note=form.note.data)
        submitter = mo.submit.Submitter(instance_path=app.instance_path)

        try:
            submitter.submit_paper(paper, tmp_path)
        except mo.submit.SubmitException as e:
            flash(f'Chyba: {e}', 'danger')
            # FIXME: Tady nemažeme tmpfile, zatím si ho chceme nechat pro analýzu.
            return redirect(url_for('user_contest_task', contest_id=contest_id, task_id=task_id))

        sess.add(paper)

        # FIXME: Bylo by hezké použít INSERT ... ON CONFLICT UPDATE
        # (SQLAlchemy to umí, ale ne přes ORM, jen core rozhraním)
        sol = (sess.query(db.Solution)
               .filter_by(task=task, user=g.user)
               .with_for_update()
               .one_or_none())
        if sol is None:
            sol = db.Solution(task=task, user=g.user)
            sess.add(sol)
        sol.final_submit_obj = paper

        sess.commit()

        flash('Řešení odevzdáno', 'success')
        return redirect(url_for('user_contest', id=contest_id))

    papers = (sess.query(db.Paper)
              .filter_by(for_user_obj=g.user, task=task, type=db.PaperType.solution)
              .options(joinedload(db.Paper.uploaded_by_obj))
              .order_by(db.Paper.uploaded_at.desc())
              .all())

    return render_template(
        'user_contest_task.html',
        contest=contest,
        task=task,
        papers=papers,
        form=form,
    )


@app.route('/user/paper/<int:id>/')
def user_paper(id: int):
    sess = db.get_session()

    paper = (sess.query(db.Paper)
             .options(joinedload(db.Paper.task).joinedload(db.Task.round))
             .get(id))

    # XXX: Tímhle dáváme útočníkům orákulum na zjišťování validity IDček,
    # ale to nevadí, protože IDčka stejně přidělujeme sekvenčně.
    if paper is None:
        return werkzeug.exceptions.NotFound()

    task = paper.task
    round = task.round

    if paper.for_user != g.user.user_id:
        logger.warn(f'Účastník #{g.user.user_id} chce cizí papír')
        return werkzeug.exceptions.Forbidden()

    if paper.type == db.PaperType.solution:
        allowed_states = [db.RoundState.running, db.RoundState.grading, db.RoundState.closed]
    elif paper.type == db.PaperType.feedback:
        allowed_states = [db.RoundState.closed]
    else:
        assert False

    if round.state not in allowed_states:
        return werkzeug.exceptions.Forbidden()

    return mo.web.util.send_task_paper(paper)
