from flask import Response, send_file
from flask_wtf import FlaskForm
import os
from sqlalchemy.orm.query import Query
from typing import Tuple
import werkzeug.exceptions
import werkzeug.utils
import wtforms

import mo.db as db
import mo.jobs
from mo.util import logger
from mo.web import app


class PagerForm(FlaskForm):
    limit = wtforms.IntegerField()
    offset = wtforms.IntegerField()
    next = wtforms.SubmitField("Další")
    previous = wtforms.SubmitField("Předchozí")

    def apply_limits(self, query: Query, pagesize: int = 50) -> Tuple[int, Query]:
        count = query.count()
        if not self.offset.data:
            self.offset.data = 0
        if not self.limit.data:
            self.limit.data = pagesize

        if self.previous.data:
            self.offset.data = max(0, self.offset.data - pagesize)
        if self.next.data:
            self.offset.data = min(count // pagesize * pagesize, self.offset.data + pagesize)

        query = query.offset(self.offset.data)
        query = query.limit(self.limit.data)

        return (count, query)


def send_task_statement(round: db.Round) -> Response:
    assert round.tasks_file is not None
    file = os.path.join(app.instance_path, 'statements', round.tasks_file)
    if os.path.isfile(file):
        return send_file(file,
                         mimetype='application/pdf',
                         attachment_filename='zadani.pdf')
    else:
        logger.error(f'Zadání {file} je v DB, ale soubor neexistuje')
        raise werkzeug.exceptions.NotFound()


def task_paper_filename(paper: db.Paper) -> str:
    """Doporučujeme preloadovat task."""

    task = paper.task

    secure_task_code = werkzeug.utils.secure_filename(task.code)
    if paper.type == db.PaperType.solution:
        return f'{secure_task_code}-reseni-{paper.paper_id}.pdf'
    elif paper.type == db.PaperType.feedback:
        return f'{secure_task_code}-opravene-{paper.paper_id}.pdf'
    else:
        assert False


def send_task_paper(paper: db.Paper) -> Response:
    file = os.path.join(app.instance_path, 'submits', paper.file_name)

    # FIXME: Předání jména funguje jen s as_attachment=True
    if os.path.isfile(file):
        return send_file(file,
                         mimetype='application/pdf')
    else:
        logger.error(f'Soubor {file} je v papers, ale ve FS neexistuje')
        raise werkzeug.exceptions.NotFound()


def send_job_result(job: db.Job) -> Response:
    assert job.out_file is not None
    file = mo.jobs.job_file_path(job.out_file)

    if file.endswith('.zip'):
        type = 'application/zip'
    else:
        type = 'application/binary'

    if os.path.isfile(file):
        return send_file(file,
                         mimetype=type,
                         as_attachment=True,
                         attachment_filename=job.out_file)
    else:
        logger.error(f'Soubor {file} je výsledkem jobu, ale ve FS neexistuje')
        raise werkzeug.exceptions.NotFound()
